/* oqtbroadcaster.h
 *
 * Part of OpenQuicktime Broadcaster
 * Copyright (C) 2002 University of Southampton
 *
 * Makes use of OpenQuicktime, a free QuickTime library.
 * Based loosly on rtp.c from the LAME 3.70 sources.
 *
 * OpenQuicktime Broadcaster is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime Broadcaster is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Nicholas Humfrey <njh@ecs.soton.ac.uk>
 * http://www.openquicktime.org/broadcaster/
 */


// oqtb_memfile.c
/* extern oqt_t* oqtb_open_memfile(int file_size);
extern char* oqtb_memfile_get_ptr(oqt_t *file);
extern int oqtb_memfile_get_max_len(oqt_t *file); */

// stsdtable.c (Part of OpenQuicktime library)
extern void oqt_write_stsd_table(oqt_t *file, oqt_minf_t *minf, oqt_stsd_table_t *table);



// Internal settings
struct global_settings {
	int		verbose_on;
	int		rtp_on;
	int		libshout_on;
	int		background_on;
	char	*record_file;
	
	char	*audio_source;
	int		audio_samplerate;
	int		audio_bits;
	int		audio_channels;
	
	char	*rtp_ip;
	int		rtp_port;
	int		rtp_ttl;
	int		rtp_samp_desc_int;
	int		rtp_packet_size_limit;
	int		rtp_packet_dur_limit;
	int		rtp_stream_id;
	
	char	codec[4];
};

struct codec_param {
	char*	name;
	char*	value;
};




/*
 * System endianness -- determined by autoconf.
 */
#ifdef WORDS_BIGENDIAN
#define RTP_BIG_ENDIAN 1
#else
#define RTP_LITTLE_ENDIAN 1
#endif


/*
 * Current protocol version.
 */
#define RTP_VERSION    2


/*
 * RTP data header
 */
typedef struct {
#if RTP_BIG_ENDIAN
    oqt_uint32_t v:2;  		  /* protocol version */
    oqt_uint32_t p:1;         /* padding flag */
    oqt_uint32_t x:1;         /* header extension flag */
    oqt_uint32_t cc:4;        /* CSRC count */
    oqt_uint32_t m:1;         /* marker bit */
    oqt_uint32_t pt:7;        /* payload type */
#elif RTP_LITTLE_ENDIAN
    oqt_uint32_t cc:4;        /* CSRC count */
    oqt_uint32_t x:1;         /* header extension flag */
    oqt_uint32_t p:1;         /* padding flag */
    oqt_uint32_t v:2;         /* protocol version */
    oqt_uint32_t pt:7;        /* payload type */
    oqt_uint32_t m:1;         /* marker bit */
#else
#error Define one of RTP_LITTLE_ENDIAN or RTP_BIG_ENDIAN
#endif
    oqt_uint32_t seq:16;      /* sequence number */
    oqt_uint32_t ts;               /* timestamp */
    oqt_uint32_t ssrc;             /* synchronization source */
   // oqt_uint32_t csrc[1];          /* optional CSRC list */
} rtp_hdr_t;







typedef struct {
#if RTP_BIG_ENDIAN
    oqt_uint32_t ver:4;   /* protocol version */
    oqt_uint32_t pck:2;   /* packetization scheme */
    oqt_uint32_t s:1;  	  /* sync sample */
    oqt_uint32_t q:1;     /* payload desc included */
    
    oqt_uint32_t l:1;     /* packet specified QT header */
    oqt_uint32_t res:7;   /* reserved bits */
    
//    unsigned int d:1;     /* cache Payload ID ? */
//    unsigned int id:15;   /* Payload ID */
    oqt_uint16_t id;   /* Payload ID*/

    //oqt_uint32_t k:1;     /* samples are sync */
   //oqt_uint32_t f:1;     /* samples are sparse */
    //oqt_uint32_t a:1;     /* contains payload desc */
    //oqt_uint32_t z:1;     /* end of payload desc */
    //oqt_uint32_t res2:4;   /* reserved bits */
    //oqt_uint32_t res3:8;   /* reserved bits */
    
    oqt_uint16_t kfaz;   /* The above */
    
	
#elif RTP_LITTLE_ENDIAN

    oqt_uint32_t q:1;     /* payload desc included */
    oqt_uint32_t s:1;  	  /* sync sample */
    oqt_uint32_t pck:2;   /* packetization scheme */
    oqt_uint32_t ver:4;   /* protocol version */

    oqt_uint32_t res:7;   /* reserved bits */
    oqt_uint32_t l:1;     /* packet specified QT header */

//    oqt_uint32_t id:15;   /* Payload ID character 1*/
//    oqt_uint32_t d:1;     /* cache Payload ID ? */
    oqt_uint16_t id;   /* Payload ID*/

//    oqt_uint32_t res2:4;   /* reserved bits */
//    oqt_uint32_t z:1;     /* end of payload desc */
//    oqt_uint32_t a:1;     /* contains payload desc */
//    oqt_uint32_t f:1;     /* samples are sparse */
//    oqt_uint32_t k:1;     /* samples are sync */
//    oqt_uint32_t res3:8;   /* reserved bits */

    oqt_uint16_t kfaz;   /* The above */

#else
#error Define one of RTP_LITTLE_ENDIAN or RTP_BIG_ENDIAN
#endif
    
    oqt_uint16_t dlen;   /* Payload description length */
    oqt_uint32_t mtype;    /* Media type */
    oqt_uint32_t timescale; /* Time scale */
    
    // 'ti' TLV
    oqt_uint16_t		ti_len;
    oqt_uint16_t		ti_type;
    oqt_uint32_t		ti_value;
    
    // 'sd' TLV
    oqt_uint16_t		sd_len;
    oqt_uint16_t		sd_type;
    oqt_uint32_t		sd_value;		// This is just used to create a pointer
} qt_head_t;


typedef struct {
	rtp_hdr_t r;
	qt_head_t q;
} oqtb_head_t;


typedef struct {
	oqt_uint16_t len;
	oqt_uint16_t type;
} qt_tlv_t;
