/* oqtimport_avi.c
 * Copyright (C) 2002 OpenQuicktime Team
 *
 * This file is part of OpenQuicktime, a free QuickTime library.
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


/* An AVI to MOV converter (originally oqtavi2mov). */



/* IMPORTANT: sound is broken!! */


/* TODO:

  - FIX THE SOUND SUPPORT
    * setup extra audio information in the QT file
      maybee it should be done on a per-audio codec basis ?
    * find the correct prototype & implementation for oqt_write_audio_data

  - support 64-bit seek to convert big avi files

  - support AVI 2.0 (OpenDML) 
  - support Vidomi Enhanced AVI (subtitles, chapter names)

*/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>

#include <openquicktime/openquicktime.h>
#include <openquicktime/structs.h>
#include <openquicktime/colormodels.h>

int debug = 0;

#define DEBUG_MSG(s) if (debug) { s; fflush(stdout); }

oqt_t* qtfile;


/* AVI */
/* --- */


oqt_t* avifile;

typedef struct {
  char         id[5];     /* ID (or FormType for LIST and RIFF chunks) */
  int          is_parent; /* 1 for LIST and RIFF chunks */
  oqt_uint32_t size;      /* size of the data (not including header) */
  oqt_uint64_t data;      /* position of the data (or first sub-chunk) */
  oqt_uint64_t header;    /* position of the header of the chunk */
  oqt_uint64_t next;      /* position of the chunk following this one */
} chunk_t;

chunk_t riff; /* big RIFF chunk */

int width, height;
int nb_streams;
int nb_frames;
double frame_rate;

#define MAX_STREAMS 20

struct {
  int  id;        /* map to which AVI stream number ? */
  int  track;     /* OQT track number */
  char handler[5];
  int  width;
  int  height;
  int  bpp;
  char format[5];
  oqt_uint64_t* key; /* list of offset of keyframes from begining of file */
  int nb_key;
  int max_key; /* size of the key buffer */
  int cur_key; /* used internally in copy_chunk */
} video_stream[MAX_STREAMS];

struct {
  int  id;        /* map to which AVI stream number ? */
  int  track;     /* OQT track number */
  char handler[5];
  char format[5];
  int  nb_channel;
  int  rate;
  int  bits;
} audio_stream[MAX_STREAMS];

/* map AVI stream number => video_stream of audio_stream entry */
int avi_stream[MAX_STREAMS];


/* mapping MS format => FOURCC for a few codecs
   you should also give the classical bit-size of sampels (when AVI set it
   to 0)
   TODO: improve this!
*/
struct {
  int   format;
  char  fourcc[5];
  int bits;
} audio_formats[] =
  { { 0x0055, { 'm','s',0,0x55}, 16 } ,
    { 0x0001, "sowt", 16 },
    { 0x0006, "alaw", 8 /* FIXME ? */ },
    { 0x0007, "ulaw", 8 /* FIXME ? */ },
    { 0x0031, "agsm", 0 /* FIXME ? */ }
  };


/* map AVI informations => QT informations */
/* TODO: actually use this table somwehere to generate INFO atoms */
struct {
  char avi[5];
  char qt[5];
  char* description;
} info[] =
  { { "IARL", "", "Archival Location" },
    { "IART", "@prf", "Artist" },
    { "ICMS", "@prd", "Commisioned" },
    { "ICMT", "@ed1", "Comments" },
    { "ICOP", "@cpy", "Copyright" },
    { "ICRD", "@day", "Creation Date" },
    { "ICRP", "", "Cropped" },
    { "IDIM", "", "Dimensions" },
    { "IDPI", "", "Dots Per Inch" },
    { "IENG", "@dir", "Engineer" }, /* director => engineer... */
    { "IGNR", "", "Genre" },
    { "IKEY", "", "Keywords" },
    { "ILGT", "", "Lightness" },
    { "IMED", "@fmt", "Medium" },
    { "INAM", "name", "Name" },
    { "IPLT", "", "Palette Setting" },
    { "IPRD", "", "Product" },
    { "ISBJ", "@inf", "Subject" },
    { "ISFT", "", "Software" },
    { "ISHP", "", "Sharpness" },
    { "ISRC", "@src", "Source" },
    { "ISRF", "", "Source Form" },
    { "ITCH", "@wrt", "Technician" } /* writer => technician... */
  };
 // Also see openquicktime/structs.h for OQT_INFO_* defines there


int nb_audio = 0;
int nb_video = 0;


static inline void avi_fread(void* buf,
			     oqt_int64_t s)
{
  if (!oqt_read_data(avifile, buf, s)) {
    fprintf(stderr,"avi_fread: Read error in AVI file.\n");
    exit(1);
  }
}

static inline oqt_uint16_t avi_word(void)
{
  unsigned char size[2];
  avi_fread(size,2);
  return (oqt_uint16_t)size[0] + ((oqt_uint16_t)size[1]<<8);
}

static inline oqt_uint32_t avi_dword(void)
{
  unsigned char size[4];
  avi_fread(size,4);
  return (oqt_uint32_t)size[0] + ((oqt_uint32_t)size[1]<<8) + ((oqt_uint32_t)size[2]<<16) + ((oqt_uint32_t)size[3]<<24);
}

static inline void avi_fourcc(char* dst)
{
  avi_fread(dst,4);
  dst[4] = 0;
}

static void avi_get_chunk (chunk_t* dst)
{
//  dst->header = ftell(avifile);
  dst->header = oqt_get_position(avifile);
  avi_fourcc(dst->id);
  DEBUG_MSG(printf("New chunk id='%s' ",dst->id));
  dst->size = avi_dword();
  if (!strcmp(dst->id,"RIFF") || !strcmp(dst->id,"LIST")) {
    avi_fread(dst->id,4);
    dst->is_parent = 1;
    dst->size-=4;
    DEBUG_MSG(printf("formtype='%s' ",dst->id));
  }
  else dst->is_parent = 0;
//  dst->data = ftell(avifile);
  dst->data = oqt_get_position(avifile);
  dst->next = dst->data + ((dst->size+1)/2)*2;
  DEBUG_MSG(printf("size=%lu header=%lu data=%lu next=%lu\n",
	       (unsigned long)dst->size,
	       (unsigned long)dst->header,
	       (unsigned long)dst->data,
	       (unsigned long)dst->next));
}

static int avi_find_chunk(const chunk_t* parent,
			  const chunk_t* prec,
			  chunk_t*       son,
			  char*          id)
{
  oqt_uint64_t pos;
  chunk_t dummy;
  if (!son) son = &dummy;
  if (prec) pos = prec->data + prec->size;
  else pos = parent->data;
  while (pos<parent->next) {
    //fseek(avifile,pos,SEEK_SET);
    oqt_set_position(avifile, pos);
    avi_get_chunk(son);
    if (!strcmp(son->id,id)) return 1;
    pos = son->next;
  }
  return 0;
}

/* gives the entry in video_stream */
static inline int avi_video_stream_of_ckid(char* id)
{
  if ((id[2]=='d' && id[3]=='b') ||
      (id[2]=='d' && id[3]=='c')) {
    int a = (id[0]-'0')*10 + (id[1]-'0');
    if (video_stream[avi_stream[a]].id==a) return avi_stream[a];
  }
  return -1;
}

/* gives the entry in audio_stream */
static inline int avi_audio_stream_of_ckid(char* id)
{
  if (id[2]=='w' && id[3]=='b') {
    int a = (id[0]-'0')*10 + (id[1]-'0');
    if (audio_stream[avi_stream[a]].id==a) return avi_stream[a];
  }
  return -1;
}

static void avi_open(void)
{
  char id[5] = "1234";
  chunk_t hdrl, avih, strl, idx1, movi;
  int i, delay;
  avi_fread(id,4);
  //rewind(avifile);
  oqt_set_position(avifile, 0);
  avi_get_chunk(&riff);
  if (strcmp(id,"RIFF") || strcmp(riff.id,"AVI ")) {
    fprintf(stderr,"Not an AVI file: invalid signature.\n");
    exit(1);
  }
  //fseek(avifile,0,SEEK_END);
  //if (!riff.size) riff.size = ftell(avifile)-riff.data;
  if (!riff.size) riff.size = oqt_get_length(avifile)-riff.data;
  

  /* find & read avih */
  if (!avi_find_chunk(&riff,NULL,&hdrl,"hdrl")) {
    fprintf(stderr,"Not an AVI file: cannot find hdrl chunk.\n");
    exit(1);
  }
  if (!avi_find_chunk(&hdrl,NULL,&avih,"avih")) {
    fprintf(stderr,"Not an AVI file: cannot find avih chunk.\n");
    exit(1);
  }
  delay = avi_dword();                  /* DWORD dwMicroSecPerFrame */
  frame_rate = 1000000./(double)delay;
  avi_dword();                          /* DWORD dwMaxBytesPerSec */
  avi_dword();                          /* DWORD dwReserved1 */
  avi_dword();                          /* DWORD dwFlags */
  nb_frames = avi_dword();              /* DWORD dwTotalFrames */
  avi_dword();                          /* DWORD dwInitialFrames */
  nb_streams = avi_dword();             /* DWORD dwStreams */
  avi_dword();                          /* DWORD dwSuggestedBufferSize */
  width = avi_dword();                  /* DWORD dwWidth */
  height = avi_dword();                 /* DWORD dwHeight */

  if (nb_streams<1 || nb_streams>=MAX_STREAMS) {
    fprintf(stderr,"Invalid number of streams %i (should be in 1-%i)!\n",
	    nb_streams, MAX_STREAMS-1);
    exit(1);
  }
  printf("oqtimport_avi: size=%ix%i, %i stream(s), %i frame(s), %g fps.\n",
	 width, height, nb_streams, nb_frames, frame_rate);
  
  /* find & read all strl */
  for (i=0;i<nb_streams;i++) {
    chunk_t strh, strl, strf;
    char fcc[5];
    if (!avi_find_chunk(&hdrl,(i==0)?NULL:&strl,&strl,"strl")) {
      fprintf(stderr,"Not an AVI file,: cannot find strl chunk for stream %i (in 0-%i).\n",i,nb_streams-1);
      exit(1);
    }
    
    /* find & read strh */
    if (!avi_find_chunk(&strl,NULL,&strh,"strh")) {
      fprintf(stderr,"Not an AVI file,: cannot find strh chunk for stream %i (in 0-%i).\n",i,nb_streams-1);
      exit(1);
    }
    avi_fourcc(fcc); /* FOURCC fccType */

    if (!strcmp(fcc,"vids")) {      

      /* video track */

      video_stream[nb_video].id = i;
      avi_stream[i] = nb_video;
      avi_fourcc(video_stream[nb_video].handler); /* FOURCC fccHandler */
      avi_dword(); /* DWORD  dwFlags */
      avi_dword(); /* DWORD  dwPriority */
      avi_dword(); /* DWORD  dwInitialFrames */
      avi_dword(); /* DWORD  dwScale */
      avi_dword(); /* DWORD  dwRate */
      avi_dword(); /* DWORD  dwStart */
      avi_dword(); /* DWORD  dwLength */
      avi_dword(); /* DWORD  dwSuggestedBufferSize */
      avi_dword(); /* DWORD  dwQuality */
      avi_dword(); /* DWORD  dwSampleSize */
      avi_dword(); /* RECT   rcFrame */

      if (!avi_find_chunk(&strl,NULL,&strf,"strf")) {
	fprintf(stderr,"Not an AVI file,: cannot find strf chunk for stream %i (in 0-%i).\n",i,nb_streams-1);
      exit(1);
      }

      /* BITMAPINFOHEADER */
      avi_dword(); /* DWORD  biSize */
      video_stream[nb_video].width = avi_dword();    /* LONG   biWidth */
      video_stream[nb_video].height = avi_dword();   /* LONG   biHeight */
      avi_word();                                    /* WORD   biPlanes */
      video_stream[nb_video].bpp = avi_word();       /* WORD   biBitCount */
      avi_fourcc(video_stream[nb_video].format);     /* DWORD  biCompression */
      avi_dword(); /* DWORD  biSizeImage */
      avi_dword(); /* LONG   biXPelsPerMeter */
      avi_dword(); /* LONG   biYPelsPerMeter */
      avi_dword(); /* DWORD  biClrUsed */
      avi_dword(); /* DWORD  biClrImportant */

      video_stream[nb_video].nb_key = 0;
      video_stream[nb_video].cur_key = 0;
      video_stream[nb_video].max_key = 100;
      video_stream[nb_video].key = (oqt_uint64_t*)malloc(sizeof(oqt_uint64_t)*video_stream[nb_video].max_key);
      
      if (avi_find_chunk(&strl,NULL,NULL,"strd"))
	printf("oqtimport_avi: codec-specific information for this stream will be discarded.\n");
      
      printf("oqtimport_avi: stream %i is video, handler='%s', format='%s', size=%ix%i, bpp=%i.\n",
	     i,
	     video_stream[nb_video].handler,
	     video_stream[nb_video].format,
	     video_stream[nb_video].width,
	     video_stream[nb_video].height,
	     video_stream[nb_video].bpp);
      nb_video++;
    } 

    else if (!strcmp(fcc,"auds")) {

      /* audio track */

      int format,j;

      audio_stream[nb_audio].id = i;
      avi_stream[i] = nb_audio;
      avi_fourcc(audio_stream[nb_audio].handler); /* FOURCC fccHandler */
      avi_dword(); /* DWORD  dwFlags */
      avi_dword(); /* DWORD  dwPriority */
      avi_dword(); /* DWORD  dwInitialFrames */
      avi_dword(); /* DWORD  dwScale */
      avi_dword(); /* DWORD  dwRate */
      avi_dword(); /* DWORD  dwStart */
      avi_dword(); /* DWORD  dwLength */
      avi_dword(); /* DWORD  dwSuggestedBufferSize */
      avi_dword(); /* DWORD  dwQuality */
      avi_dword(); /* DWORD  dwSampleSize */
      avi_dword(); /* RECT   rcFrame */
 

      if (!avi_find_chunk(&strl,NULL,&strf,"strf")) {
	fprintf(stderr,"Not an AVI file,: cannot find strf chunk for stream %i (in 0-%i).\n",i,nb_streams-1);
      exit(1);
      }

      /* WAVEFORMATEX */

      /* Remark: this sems to be the same as for .WAV files! */

      format = avi_word();                            /* WORD  wFormatTag */   
      audio_stream[nb_audio].nb_channel = avi_word(); /* WORD  nChannels */
      audio_stream[nb_audio].rate = avi_dword(); /* DWORD nSamplesPerSec */
      avi_dword();                               /* DWORD nAvgBytesPerSec */
      avi_word();                                /* WORD  nBlockAlign */
      audio_stream[nb_audio].bits = avi_word();  /* WORD  wBitsPerSample */  
      avi_word();                                /* WORD  cbSize */

      for (j=0;j<sizeof(audio_formats)/sizeof(audio_formats[0]);j++)
	if (audio_formats[j].format==format) break;

      if (j<sizeof(audio_formats)/sizeof(audio_formats[0])) {

	memcpy(audio_stream[nb_audio].format, audio_formats[j].fourcc, 5);

	if (!audio_stream[nb_audio].bits)
	  audio_stream[nb_audio].bits = audio_formats[j].bits;

	printf("oqtimport_avi: stream %i is audio, handler='%s', format='%s', %i sps, %i channel(s), %i bits.\n",i,
	       audio_stream[nb_audio].handler,
	       audio_stream[nb_audio].format,
	       audio_stream[nb_audio].rate,
	       audio_stream[nb_audio].nb_channel,
	       audio_stream[nb_audio].bits);
	nb_audio++;
      }
      else
	printf("oqtimport_avi: stream %i has unknown format 0x%04X, skipped.\n",i,
	       format);
    }

    else {
      avi_stream[i] = -1;
      printf("oqtimport_avi: stream %i is ignored (type='%s').\n",i,fcc);
    }
  }

  /* find movi */
  if (!avi_find_chunk(&riff,NULL,&movi,"movi")) {
    fprintf(stderr,"Not an AVI file,: cannot find mov chunk.\n");;
    exit(1);
  }
  
 /* find idx1 & get keyframes */
  if (avi_find_chunk(&riff,NULL,&idx1,"idx1")) {
    oqt_uint64_t i, nb = idx1.size / 16, pos;
    char id[5];
    int flags, v;
    
    for (i=0;i<nb;i++) {
      avi_fourcc(id);      /* DWORD ckid */
      flags = avi_dword(); /* DWORD dwFlags */
      pos = avi_dword();   /* DWORD dwChunkOffset */
      avi_dword();         /* DWORD dwChunkLength */
      
      v = avi_video_stream_of_ckid(id);
      if (v>=0 && (flags & 0x10)) {
	video_stream[v].key[video_stream[v].nb_key] = movi.data - 4 + pos;
	video_stream[v].nb_key++;
	if (video_stream[v].nb_key==video_stream[v].max_key) {
	  video_stream[v].max_key *= 2;
	  video_stream[v].key = realloc(video_stream[v].key,
					video_stream[v].max_key * 
					sizeof(oqt_uint64_t));
	}
      }
    }

    for (i=0;i<nb_video;i++)
      printf("oqtimport_avi: video stream %i has %i keyframe(s).\n",
	     (int)i, video_stream[i].nb_key);

  }
  else printf("oqtimport_avi: no idx1 chunk => no keyframe.\n");
}


/* AVI => QT */
/* --------- */

static void create_tracks(void)
{
  int i;

  /* video */
  for (i=0;i<nb_video;i++) {
    video_stream[i].track = oqt_add_video_track(qtfile, 
					       video_stream[i].width, 
					       video_stream[i].height,
					       frame_rate, 
					       video_stream[i].format);
    if (video_stream[i].track<0) {
      fprintf(stderr,"Cannot create video track %i.\n",i);
      exit(1);
    }

    if (!oqt_supported_video(qtfile, video_stream[i].track))
      printf("oqtimport_avi: WARNING, OQT does not seem to understand video codec '%s', movie may be unplayable.\n",video_stream[i].format);
    
    oqt_set_video_depth(qtfile,video_stream[i].track,video_stream[i].bpp);
  }  

  /* audio */
  for (i=0;i<nb_audio;i++) {
    audio_stream[i].track = oqt_add_audio_track(qtfile, 
						audio_stream[i].nb_channel,
						audio_stream[i].rate,
						audio_stream[i].bits,
						audio_stream[i].format);
    if (audio_stream[i].track<0) {
      fprintf(stderr,"Cannot create audio track %i.\n",i);
      exit(1);
    }
    
    if (!oqt_supported_audio(qtfile, audio_stream[i].track))
      printf("oqtimport_avi: WARNING, OQT does not seem to understand audio codec '%s', movie may be unplayable.\n",audio_stream[i].format);

    /* FIXME:
       do we need to pass informations to OQT directly in tables
       as it it done in audioplugin/mp3/codec_mp3.c for instance ?
       are these informations specific to each audio codec
     */
  } 
}


static void copy_chunk(chunk_t* cur)
{
  int v;
  static oqt_uint8_t* buf = NULL;
  static size_t size = 0;
  
  v = avi_audio_stream_of_ckid(cur->id);
  if (v>=0) {
    /* audio chunk */
    if (cur->size>size) {
      size = cur->size*2;
      buf = realloc(buf, size);
    }
    avi_fread(buf,cur->size);   

    /* FIXME: 
       correct the oqt_write_audio_data mechanism here and in OQT lib
       how can we know the number of samples corresponding to this chunk ?
    */
    oqt_write_audio_data(qtfile, audio_stream[v].track, buf, cur->size, 1152, 1, NULL);
  }
  
  v = avi_video_stream_of_ckid(cur->id);
  if (v>=0) {
    /* video chunk */
    int key;
    if (cur->size>size) {
      size = cur->size*2;
      buf = realloc(buf, size);
    }
    while (video_stream[v].cur_key < video_stream[v].nb_key &&
	   video_stream[v].key[video_stream[v].cur_key] < cur->header)
      video_stream[v].cur_key++;

    key = video_stream[v].cur_key < video_stream[v].nb_key &&
      video_stream[v].key[video_stream[v].cur_key] == cur->header;
    
    avi_fread(buf,cur->size);
    oqt_write_video_data(qtfile, video_stream[v].track, buf, cur->size, key);
  }
}

/* we copy the chunks in the order they appear in the AVI file, independently
   of the track number
   the result is that video and audio are interleaved in the QT file if it
   was the case in the AVi file
*/
static void copy(void)
{
  chunk_t movi;
  chunk_t cur, cur2;
  oqt_uint64_t pos;
 
  /* find & traverse movi */
  if (!avi_find_chunk(&riff,NULL,&movi,"movi")) {
    fprintf(stderr,"Not an AVI file,: cannot find mov chunk.\n");;
    exit(1);
  }
  
  pos = movi.data;
  while (pos<movi.next) {
//    fseek(avifile,pos,SEEK_SET);
	oqt_set_position(avifile, pos);
    avi_get_chunk(&cur);

    if (!strcmp(cur.id,"rec")) {
      pos = movi.data;
      while (pos<cur.next) {
//	fseek(avifile,pos,SEEK_SET);
	oqt_set_position(avifile, pos);
	avi_get_chunk(&cur2);
	copy_chunk(&cur2);
	pos = cur2.next;
      }
    }
    else copy_chunk(&cur);
    pos = cur.next;
  }

}




/* main */
/* ---- */

static void help(void)
{
	printf("oqtimport_avi [options] infile outfile\n");
	printf("options: --debug\n");
	exit(1);
}

int main(int argc, char** argv)
{
	char* infilename = NULL;  
	char* outfilename = NULL;
	
	/* parse arguments */
	
	for (argc--,argv++;argc && *argv;argc--,argv++) {
	if (!strcmp(*argv,"--debug")) debug = 1;
	else if (!strcmp(*argv,"-?") || !strcmp(*argv,"-h") || !strcmp(*argv,"--help") || **argv=='-') help();
	else if (!infilename) infilename = *argv;
	else outfilename = *argv;
	}
	
	if (!infilename || !outfilename) help();
	
	/* open AVI */
	
	//avifile = fopen(infilename,"rb");
	avifile = oqt_open(infilename);
	if (!avifile) {
	fprintf(stderr,"Cannot find AVI source file '%s'.\n",infilename);
	exit(1);
	}
	avi_open();
	
	
	/* create QT */
	
	qtfile = oqt_open_file(outfilename,OQT_MODE_WRITE);
	if (!qtfile) {
	fprintf(stderr,"Cannot create file QT destination file '%s'.\n",
		outfilename);
	exit(1);
	}
	
	/***** Add Info to movie stating how movie was made and source AVI *****/
	
	oqt_write_moviedata_init(qtfile);
	create_tracks();
	copy();
	oqt_write_moviedata_end(qtfile);
	
	oqt_close(qtfile);
	oqt_close(avifile);
	
	return 0;
}
